"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import logging

logging.basicConfig()
logger = logging.getLogger('MISC')    

def evaluate_expression(expr,variables):
  """ Evaluate an arithmetic expression "expr" written in python compatible format, using the dictionary "variables" for the values of the variables."""
  expr = expr.replace('true','True')
  expr = expr.replace('false','False')
  code = compile(expr.strip(),'<expr>','eval')
  try:
    return eval(code,dict(variables))	# important: it is necessary to copy variables, evne if it is already a dictionary because it is modified by the function eval.
  except Exception as e:
    return None
    
    
def split_constraints(constraint):
  """ Split a guard (or an invariant) into a conjunction of constraints with balanced parenthesis
  """
  constraint = constraint.replace('&&','and')
  constraint = constraint.replace('&','and')
  tokens = constraint.split("and")
  constraints = []
  while tokens:
    token = tokens.pop(0)
    if token.count("(") != token.count(")"):
      if tokens:
	tokens[0] = token + tokens[0]
      else:
	logger.error("Unbalanced parenthesis in constraint: \"" + constraint + "\"")
    else:
      constraints += [token]
  return constraints

