import logging

import ppl_ta_int
import copy

logging.basicConfig(level=logging.DEBUG)
my_logger = logging.getLogger('Python-PPL-TA')

from  ppl_ta_int import Constraint

class Variable:
  def __init__(self,i):
    self._var = ppl_ta_int.Variable(i)
  
  def __str__(self):
    return ParamFederation.context.formatExpression(self._var.__str__())
  __repr__ = __str__
  
  def id(self):
    return self._var.id()
  
  def __add__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) + arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) + LinearExpression(arg)
  
  def __sub__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) - arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) - LinearExpression(arg)
    
  def __eq__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) == arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) == LinearExpression(arg)
  
  def __le__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) <= arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) <= LinearExpression(arg)
      
  def __ge__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) >= arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) >= LinearExpression(arg)
      
  def __lt__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) < arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) < LinearExpression(arg)
      
  def __gt__(self,arg):
    if isinstance(arg, LinearExpression):
      return LinearExpression(self) > arg
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return LinearExpression(self) > LinearExpression(arg)

    
class LinearExpression:
  def __init__(self,arg=None):
    if not arg:
      self._expr = ppl_ta_int.Linear_Expression()
    elif isinstance(arg, ppl_ta_int.Linear_Expression):
      self._expr = arg
    elif isinstance(arg,Constraint):
      self._expr = ppl_ta_int.Linear_Expression(arg)
    elif isinstance(arg,Variable):
      self._expr = ppl_ta_int.Linear_Expression(arg._var)
    else:
      assert(isinstance(arg,int))
      self._expr = ppl_ta_int.Linear_Expression(arg)
  
  def copy(self):
    ret = self.__class__()
    ret._expr = ppl_ta_int.Linear_Expression(self._expr)
    return ret
    __deepcopy__ = lambda self, memo: self.copy()
    __copy__ = copy
  
  def __str__(self):
    return ParamFederation.context.formatExpression(self._expr.__str__())
  __repr__ = __str__
  
  def space_dimension():
    return self._expr.space_dimension()
  
  def is_zero():
    return self._expr.is_zero()
  
  def all_homogeneous_terms_are_zero():
    return self._expr.all_homogeneous_terms_are_zero()

  def __add__(self,arg):
    ret = LinearExpression()
    if isinstance(arg, LinearExpression):
      ret._expr = ppl_ta_int.LinearAdd(self._expr,arg._expr)
      return ret
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return self + LinearExpression(arg)
  
  def __radd__(self,arg):
    assert(isinstance(arg,int))
    return LinearExpression(arg) + self 
    
  def __sub__(self,arg):
    ret = LinearExpression()
    if isinstance(arg, LinearExpression):
      ret._expr = ppl_ta_int.LinearSub(self._expr,arg._expr)
      return ret
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return self - LinearExpression(arg)
  
  def __rsub__(self,arg):
    assert(isinstance(arg,int))
    return LinearExpression(arg) - self 
  
  def __mul__(self,arg):
    assert(isinstance(arg,int))
    ret = LinearExpression()
    ret._expr = ppl_ta_int.LinearMulRight(self._expr,arg)
    return ret
      
  def __rmul__(self,arg):
    assert(isinstance(arg,int))
    ret = LinearExpression()
    ret._expr = ppl_ta_int.LinearMulLeft(arg,self._expr)
    return ret
      
  def __iadd__(self,arg):
    if isinstance(arg, LinearExpression):
      ppl_ta_int.LinearIAdd(self._expr,arg._expr)
      return self
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))    
      self += LinearExpression(arg)
      return self
  
  def __isub__(self,arg):
    if isinstance(arg, LinearExpression):
      ppl_ta_int.LinearISub(self._expr,arg._expr)
      return self
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))    
      self -= LinearExpression(arg)
      return self
    
  def __imul__(self,arg):
    assert(isinstance(arg,int))
    ppl_ta_int.LinearIMul(self._expr,arg)
    return self

  def __eq__(self,arg):
    if isinstance(arg, LinearExpression):
      return ParamFederation(ppl_ta_int.ConsEq(self._expr,arg._expr))
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return ParamFederation(ppl_ta_int.ConsEq(self._expr,LinearExpression(arg)._expr))
  
  def __le__(self,arg):
    if isinstance(arg, LinearExpression):
      return ParamFederation(ppl_ta_int.ConsLeq(self._expr,arg._expr))
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return ParamFederation(ppl_ta_int.ConsLeq(self._expr,LinearExpression(arg)._expr))
      
  def __ge__(self,arg):
    if isinstance(arg, LinearExpression):
      return ParamFederation(ppl_ta_int.ConsGeq(self._expr,arg._expr))
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return ParamFederation(ppl_ta_int.ConsGeq(self._expr,LinearExpression(arg)._expr))
      
  def __lt__(self,arg):
    if isinstance(arg, LinearExpression):
      return ParamFederation(ppl_ta_int.ConsLt(self._expr,arg._expr))
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return ParamFederation(ppl_ta_int.ConsLt(self._expr,LinearExpression(arg)._expr))
      
  def __gt__(self,arg):
    if isinstance(arg, LinearExpression):
      return ParamFederation(ppl_ta_int.ConsGt(self._expr,arg._expr))
    else:
      assert(isinstance(arg,int) or isinstance(arg,Variable))
      return ParamFederation(ppl_ta_int.ConsGt(self._expr,LinearExpression(arg)._expr))
      

#class Constraint:
  #def __init__(self,arg):
    #assert(isinstance(arg, ppl_ta_int.Constraint))
    #self._cons = arg
  
  #def copy(self):
    #ret = self.__class__()
    #ret._expr = ppl_ta_int.Constraint(self._cons)
    #return ret
    #__deepcopy__ = lambda self, memo: self.copy()
    #__copy__ = copy
  
  #def space_dimension ():
    #return self._cons.space_dimension()
  
  #def is_equality ():
    #return self._cons.is_equality()
  
  #def is_inequality ():
    #return self._cons.is_inequality()
  
  #def is_nonstrict_inequality ():
    #return self._cons.is_nonstrict_inequality()
  
  #def is_strict_inequality ():
    #return self._cons.is_strict_inequality()
  
  #def is_tautological ():
    #return self._cons.is_tautological()
  
  #def is_inconsistent ():
    #return self._cons.is_inconsistent()

  #def is_equivalent_to (arg):
    #assert(isinstance(arg,Constraint))
    #return self._cons.is_equivalent_to(arg._cons)

  #def __and__(self,arg):
    #assert(isinstance(arg,Constraint))
    #ret = ParamFederation(self)
    #ret.addConjunction(arg)
    #return
  
  #def __or__(self,arg):
    #assert(isinstance(arg,Constraint))
    #ret = ParamFederation(self)
    #ret.addDisjunction(arg)
    #return

class ParamFederation:
  
  context = None

  def __init__(self,arg=None):
    if not arg:
      self._fed = ppl_ta_int.ParamFederation(len(ParamFederation.context.variables))
    else:
      assert(isinstance(arg,Constraint))
      self._fed = ppl_ta_int.ParamFederation(len(ParamFederation.context.variables))
      self._fed.add_conjunction(arg)
          
  def __str__(self):
    return ParamFederation.context.formatExpression(self._fed.toStr())
  __repr__ = __str__
  
  def copy(self):
    ret = self.__class__()
    ret._fed = ppl_ta_int.ParamFederation(self._fed)
    return ret
  __deepcopy__ = lambda self, memo: self.copy()
  __copy__ = copy
  
  def __and__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    ret = self.copy()
    ret._fed.add_conjunction(arg2._fed)
    return ret  
      
  def __iand__(self, arg2):        
    assert(isinstance(arg2, ParamFederation))
    self._fed.add_conjunction(arg2._fed)
    return self
    
  def __or__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    ret = self.copy()
    ret._fed.add_disjunction(arg2._fed)
    return ret
  def __ior__(self, arg2):        
    assert(isinstance(arg2, ParamFederation))
    self._fed.add_disjunction(arg2._fed)
    return self
  
  def __sub__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    ret = self.copy()
    ret._fed.difference(arg2._fed)
    return ret
  def __isub__(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    self._fed.difference(arg2._fed)
    return self
    
  def __invert__(self):
    ret = self.__class__()
    ret._fed = self._fed.negation()
    return ret
    
  def addDisjunction(self,g):   
    assert(isinstance(g,Constraint))
    self._fed.add_disjunction(g)
    
  def addConjunction(self,g):   
    assert(isinstance(g,Constraint))
    self._fed.add_conjunction(g)

  def up(self):
    ret = self.copy()
    ret.upInPlace()
    return ret 
  def upInPlace(self):
    self._fed.future(1)
  
  def down(self):
    ret = self.copy()
    ret.downInPlace()
    return ret 
  def downInPlace(self):
    self._fed.past(1)

  def freeVariable(self, v_id):
    ret = self.copy()
    ret.freeVariableInPlace(v_id)
    return ret 
  def freeVariableInPlace(self, v_id):
    self._fed.free(v_id)
  
  def resetValue(self, v_id):
    ret = self.copy()
    ret.resetValueInPlace(v_id)
    return ret
  
  def resetValueInPlace(self, v_id):
    self._fed.reset(v_id)
  
  def post(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    ret = self.copy()
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    ret._fed.post(guard._fed,reset_vector)
    return ret
  
  def postInPlace(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    self._fed.post(guard._fed,reset_vector)
   
  def pred(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    ret = self.copy()
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    ret._fed.pred(guard._fed,reset_vector)
    return ret
  
  def predInPlace(self,guard, reset_list):
    assert(isinstance(guard, ParamFederation))
    reset_vector = ppl_ta_int.IntVector(len(reset_list))
    for i in xrange(0,len(reset_list)):
      reset_vector[i]=reset_list[i]
    self._fed.pred(guard._fed,reset_vector)
         
  def predt(self, arg2):
    assert(isinstance(arg2, ParamFederation))
    ret = self.__class__()
    ret._fed = ppl_ta_int.ParamFederation_predt(self._fed,arg2._fed,1)
    return ret 

  def pairwiseReduce(self):
    self._fed.pairwise_reduce()
    
  def contains (self, arg2):
    return self._fed.contains(arg2._fed)
    
  def isDisjointFrom(self, arg2):
    return self._fed.is_disjoint_from(arg2._fed)
    
  def geometricallyEquals(self, arg2):
    return self._fed.geometrically_equals(arg2._fed)
  def geometricallyCovers(self, arg2):
    return self._fed.geometrically_covers(arg2._fed)
    
  def paramProjection(self):
    ret = self.copy()
    ret._fed = self._fed.param_projection(1)
    return ret
    
  def paramProjectionInPlace(self):
    self._fed.param_projection(1)
    
  def hasZeroParamDomain(self):
    zero = self.copy()
    for i in xrange(1,len(ParamFederation.context.variables)):
      zero.addConjunction(Variable(i)==0)
    zero.paramProjectionInPlace()
    return zero
    
  def minimizeParam(self):
    found,min_n,min_d,minimum = self._fed.minimize_param(0)
    if found:
      return (min_n,min_d,minimum)
    else:
      return None

  def isEmpty(self):
    return self._fed.is_empty()
  def getSize(self):
    return self._fed.size();


class ParamContext:
  def __init__(self, clock_names, param_name):
    self.variables = {}
    v_id = 0
    self.variables[param_name] = Variable(v_id)
    v_id += 1
    for clock in clock_names:
      self.variables[clock] = Variable(v_id)
      v_id += 1
    ParamFederation.context = self
       
  def getEmptyParamFederation(self):
    ret = ParamFederation()
    ret._fed = ppl_ta_int.ParamFederation(len(self.variables),ppl_ta_int.EMPTY)
    return ret
      
  def getTautologyParamFederation(self):
    return ParamFederation()

  def getInitialDomain(self,min_param_value=0,max_param_value="inf"):
    ret = ParamFederation()
    ret.addConjunction(Variable(0) >= min_param_value)
    if max_param_value < "inf":
      ret.addConjunction(Variable(0) <= max_param_value)
    for i in xrange(1,len(self.variables)):
      ret.addConjunction(Variable(i) >= 0)
    return ret
  
  def getZeroDomain(self,min_param_value=0,max_param_value="inf"):
    ret = ParamFederation()
    ret.addConjunction(Variable(0) >= min_param_value)
    if max_param_value < "inf":
      ret.addConjunction(Variable(0) <= max_param_value)
    for i in xrange(1,len(self.variables)):
      ret.addConjunction(Variable(i) == 0)
    return ret
  
  def getVariable(self,v_name):
    return self.variables[v_name]
    
  def formatExpression(self,expr):
    if len(self.variables)<=26:
      for v_name, v in self.variables.items():
	expr = expr.replace(chr(v.id()+65),v_name)
      return expr
    else:
      return expr
      
