%module ppl_ta_int

%{
#define SWIG_FILE_WITH_INIT
#include "ppl_ta_int.h"
%}

%include "std_string.i"
%include "std_vector.i"
%include "typemaps.i"

%rename(LinearAdd) operator+(const Linear_Expression &e1, const Linear_Expression &e2);
%rename(LinearSub) operator-(const Linear_Expression &e1, const Linear_Expression &e2);
%rename(LinearMulRight) operator*(const Linear_Expression &e, int n);
%rename(LinearMulLeft) operator*(int n, const Linear_Expression &e);
%rename(LinearIAdd) operator+=(Linear_Expression &e1, const Linear_Expression &e2);
%rename(LinearISub) operator-=(Linear_Expression &e1, const Linear_Expression &e2);
%rename(LinearIMul) operator*=(Linear_Expression &e, int n);
  
%rename(ConsEq) operator==(const Linear_Expression &e1, const Linear_Expression &e2);
%rename(ConsLeq) operator<=(const Linear_Expression &e1, const Linear_Expression &e2);
%rename(ConsGeq) operator>=(const Linear_Expression &e1, const Linear_Expression &e2);
%rename(ConsLt) operator<(const Linear_Expression &e1, const Linear_Expression &e2);
%rename(ConsGt) operator>(const Linear_Expression &e1, const Linear_Expression &e2);


namespace std {
   %template(IntVector) vector<int>;
}

enum Degenerate_Element {UNIVERSE,EMPTY};

class Variable
{
  public:
    Variable(int i);
    int  id();
};

%extend Variable {
  std::string __str__() {
    stringstream out;
    out << * $self;
    return out.str();
  }
};

class Linear_Expression {
  public:
    Linear_Expression();
    Linear_Expression(const Linear_Expression &e);
    Linear_Expression(int n);
    Linear_Expression(Variable v);
    Linear_Expression(const Constraint &c);

    int space_dimension() const;
    bool is_zero () const;
    bool all_homogeneous_terms_are_zero () const;
    
    friend Linear_Expression operator+(const Linear_Expression &e1, const Linear_Expression &e2);
    friend Linear_Expression operator-(const Linear_Expression &e1, const Linear_Expression &e2);
    friend Linear_Expression operator*(const Linear_Expression &e, int n);
    friend Linear_Expression operator*(int n, const Linear_Expression &e);
    friend Linear_Expression & operator+=(Linear_Expression &e1, const Linear_Expression &e2);
    friend Linear_Expression & operator-=(Linear_Expression &e1, const Linear_Expression &e2);
    friend Linear_Expression & operator*=(Linear_Expression &e, int n);
};

%extend Linear_Expression {
  std::string __str__() {
    stringstream out;
    out << * $self;
    return out.str();
  }
};

class Constraint {
  public:
    Constraint (const Constraint &c);
    int space_dimension () const;
    bool is_equality () const;
    bool is_inequality () const;
    bool is_nonstrict_inequality () const;
    bool is_strict_inequality () const;
    bool is_tautological () const;
    bool is_inconsistent () const;
    bool is_equivalent_to (const Constraint &y) const;

    friend Constraint operator==(const Linear_Expression &e1, const Linear_Expression &e2);
    friend Constraint operator<=(const Linear_Expression &e1, const Linear_Expression &e2);
    friend Constraint operator>=(const Linear_Expression &e1, const Linear_Expression &e2);
    friend Constraint operator<(const Linear_Expression &e1, const Linear_Expression &e2);
    friend Constraint operator>(const Linear_Expression &e1, const Linear_Expression &e2);
};

%extend Constraint {
  std::string __str__() {
    stringstream out;
    out << * $self;
    return out.str();
  }
};

class ParamFederation 
{
  public:
    ParamFederation(int num_dimensions=0,Degenerate_Element kind=UNIVERSE);
    ParamFederation(const ParamFederation& fed);
    
    void add_disjunction(const Constraint& c);
    void add_disjunction(const ParamFederation& fed);
    void add_conjunction(const Constraint& c);
    void add_conjunction(const ParamFederation& fed);
    ParamFederation negation() const;
    static Constraint negation(const Constraint& c);
    void difference(const ParamFederation& fed);
    void future(int num_param);
    void past(int num_param);
    void reset(int var);
    void free(int var);
    void post(const ParamFederation& guard, const std::vector<int>& reset);
    void pred(const ParamFederation& guard, const std::vector<int>& reset);
    static ParamFederation predt(const ParamFederation& good,const ParamFederation& bad,int num_param);        
    void param_projection(int num_param);
    void minimize_param(int, bool *OUTPUT, int *OUTPUT, int *OUTPUT, bool *OUTPUT);
    std::string toStr() const;
    int size() const;
    bool is_empty () const;
    bool contains (const ParamFederation& ) const;
    bool is_disjoint_from(const ParamFederation& ) const;
    bool geometrically_equals(const ParamFederation& ) const;
    bool geometrically_covers(const ParamFederation& ) const;
    void pairwise_reduce ();    
};
