"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import re
import sys
import logging
import copy

import pyppl
import pynta

logging.basicConfig()
logger = logging.getLogger('PARAM_NTA')


class ParamTemplate(pynta.Template):
  def __init__(self,template,param_name,param_value=0):
    template = copy.deepcopy(template)
    pynta.Template.__init__(self,template.nta,template.name,template.locations,template.transitions,template.init_location, \
			    template.clocks,template.constants,template.variables,template.fs,template.fa,template.fb,template.fe,template.ff)
    for l in self.locations:
      l.template = self
    for t in self.transitions:
      t.template = self			    
    self.parse_template()    
    self.param_name = param_name
    if not self.constants.has_key(self.param_name):
      self.constants[self.param_name] = param_value
    else:
      logger.warning("Template has already declared a constant \"" + self.param_name + "\"")
    self.param_context = None
    self.initial_domain = None
    
  def setParamValue(self,val):
    self.constants[self.param_name] = val
    
  def parse_param(self,param_min_value,param_max_value):
    """ Parse the parametric guards and invariants of this template.
	Requires that the dbm have been previously parsed"""
    self.param_context = pyppl.ParamContext(self.context.clocks.keys(),self.param_name)
    self.initial_domain = self.param_context.getInitialDomain(param_min_value,param_max_value)
    logger.debug("initial domain: " + str(initial_domain))
    
    for location in self.locations:
      location.invariant_ppl = parse_param_invariant_or_guard(location.invariant)
      location.invariant_ppl &= initial_domain
    for transition in self.transitions:
      transition.guard_ppl = parse_param_invariant_or_guard(transition.guard)
      transition.update_ppl = [param_context.getVariable(clock.name.split('.')[1]) for clock in transition.update_clock]	

  def parse_param_invariant_or_guard(self,constraints):
    """ Parse a parametric guard into a ParamFederation domain.
    """
    if constraints == "":
      return self.param_context.getTautologyParamFederation()	  
    param_fed = self.param_context.getTautologyParamFederation()
    for constraint in constraints.split('&&'):
      code = compile(constraint,"<param>","eval")
      try:
	val = eval(code, dict(self.nta.constants.items() + self.constants.items() + self.param_context.variables.items()))  # the parameter constant value is erased by the PPL variables from the context
	param_fed.addConjunction(val)
      except Exception as e:
	logger.error('Cannot parse parametric constraint: ' + constraint)
    return param_fed

    
    