"""PyECDAR, a open source software for timed systems.
Copyright (C) 2013 INRIA Rennes

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA."""

import sys
import logging

import pysymbolicstates

logging.basicConfig()
logger = logging.getLogger('REACH')

      
class ReachabilityAnalysis:
    def __init__(self,goal_template,goal_locations_name,safety):
        self.goal_template = goal_template
        self.context = self.goal_template.context
        self.initial_discrete_state = self.goal_template.get_initial_discrete_state()
	self.safety = safety
	
	self.zone = {}
	self.accepting = {}
       	for l in self.goal_template.locations:
	  self.accepting[l] = (l.name in goal_locations_name)
        self.witness_trace = []
    	
    def next(self,state,zone):
        ret = []
        for transition in state.location.forward:
	  next_state = transition.next_state(state)
	  if next_state:
	    next_zone = zone.post(transition)
	    if next_zone:
	      ret.append((state,zone,transition,next_state,next_zone))
	return ret
    
    def solve(self):
        if self.accepting[self.initial_discrete_state.location]:
	  return False

        self.zone = {}
        self.witness_trace = []
        waiting = []
       	history = {}
	history[self.initial_discrete_state] = []
        
	initial_zone = self.context.getZeroFederation()
        initial_zone.upInPlace()
        initial_zone &= self.initial_discrete_state.location.invariant_dbm
        initial_zone.extrapolateMaxBoundsInPlace(self.goal_template.clock_bounds)
	initial_sym_state = pysymbolicstates.SetOfSymbolicStates(self.goal_template)
	initial_sym_state += [pysymbolicstates.SymbolicState(None,initial_zone)]
	self.zone[self.initial_discrete_state] = initial_sym_state
	for e in self.next(self.initial_discrete_state,initial_sym_state):
	  logger.debug("adding to wait next initial: " + str(e[3]) + "#" + str(e[4]))
	  waiting.append(e)

	i = 0
        while waiting:
	    e = waiting.pop()
            state = e[0]
            loc = e[0].location
            zone = e[1]
            transition = e[2]
            next_state = e[3]
            next_loc = e[3].location
            next_zone = e[4]
            history[next_state] = history[state] + [transition]
            i += 1
            if not self.zone.has_key(next_state) or not next_zone.is_included_in(self.zone[next_state]): # not included
	      logger.debug(str(i)+'Exploring new state:' + str(next_state) + "#" + str(next_zone) + ' (from ' + str(transition) + ')')
	      if self.zone.has_key(next_state):
		self.zone[next_state] |= next_zone
	      else:
		self.zone[next_state] = next_zone
	      if not self.accepting[next_loc]:
		for next_e in self.next(next_state,next_zone):
		  logger.debug("adding to wait next: " + str(next_e[3]) + "#" + str(next_e[4]))
		  if not next_e in waiting: # Can the contrary happen ?
		    waiting.append(next_e)
		  else:
		    logger.debug("already in wait: " + str(next_e[3]) + "#" + str(next_e[4]))
	      else:
		# goal location reached		
		self.witness_trace = history[next_state]
		return not self.safety
	logger.info(str(i) + " states visited.")
        return self.safety

    def refine(self,delta_value):
      """ Refine the spoiling strategy to extract parametrically the minimum value of delta.
      """
      param_context = self.goal_template.param_context
      initial_domain = param_context.getInitialDomain(0,delta_value)

      if not self.witness_trace:
	logger.debug("No counterexample computed")
	return param_context.getEmptyParamFederation()
      param_zone = param_context.getZeroDomain(0,delta_value)
      param_zone.upInPlace()
      param_zone &= self.initial_discrete_state.location.invariant_ppl
      param_zone &= initial_domain
      
      for t in self.witness_trace:
	param_zone.postInPlace(t.guard_ppl,t.update_ppl)
	param_zone.upInPlace()
	param_zone &= t.target.invariant_ppl
	param_zone &= self.initial_domain
	logger.debug("current domain: " + str(param_zone))

      #print param_zone.minimizeParam()	
      return param_zone.minimizeParam()	
	
      
